@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -56,8 +56,10 @@
 #include "pool_config.h"
 #include "pool_process_context.h"
 #include "pool_process_reporting.h"
+#include "watchdog/wd_ext.h"
 
 #define MAX_FILE_LINE_LEN    512
+#define MAX_USER_PASSWD_LEN  128
 
 extern void pcp_set_timeout(long sec);
 volatile sig_atomic_t pcp_exit_request; /* non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived */
@@ -177,13 +179,6 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 		}
 
 		rsize = ntohl(rsize);
-
-		if (rsize <= 0 || rsize >= MAX_PCP_PACKET_LENGTH)
-		{
-			pool_error("pcp_child: incorrect packet length (%d)", rsize);
-			exit(1);
-		}
-
 		if ((rsize - sizeof(int)) > 0)
 		{
 			buf = (char *)malloc(rsize - sizeof(int));
@@ -231,16 +226,19 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 			{
 				int len;
 				int wsize;
-				char *msg;
+				char *msg = "OperationInProgress";
+				if(Req_info->request_queue_tail != Req_info->request_queue_head)
+				{
+					POOL_REQUEST_KIND reqkind;
+					reqkind = Req_info->request[(Req_info->request_queue_head +1) % MAX_REQUEST_QUEUE_SIZE].kind;
 
-				if (Req_info->kind == NODE_UP_REQUEST)
-					msg = "FailbackInProgress";
-				else if (Req_info->kind == NODE_DOWN_REQUEST)
-					msg = "FailoverInProgress";
-				else if (Req_info->kind == PROMOTE_NODE_REQUEST)
-					msg = "PromotionInProgress";
-				else
-					msg = "OperationInProgress";
+					if (reqkind == NODE_UP_REQUEST)
+						msg = "FailbackInProgress";
+					else if (reqkind == NODE_DOWN_REQUEST)
+						msg = "FailoverInProgress";
+					else if (reqkind == PROMOTE_NODE_REQUEST)
+						msg = "PromotionInProgress";
+				}
 
 				len = strlen(msg) + 1;
 				pcp_write(frontend, "e", 1);
@@ -419,7 +417,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 				process_list = pool_get_process_list(&process_count);
 
-                mesg = (char *)malloc(7*process_count); /* PID is at most 6 characters long */
+				mesg = (char *)malloc(6*process_count);	/* port# is at most 5 characters long (MAX:65535) */
 				if (mesg == NULL)
 				{
 					pool_error("pcp_child: malloc() failed. reason: %s", strerror(errno));
@@ -430,10 +428,10 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 				for (i = 0; i < process_count; i++)
 				{
-                    char process_id[7];
-                    snprintf(process_id, sizeof(process_id), "%d", process_list[i]);
-                    snprintf(mesg+total_port_len, strlen(process_id)+1, "%s", process_id);
-                    total_port_len += strlen(process_id)+1;
+					char port[6];
+					snprintf(port, sizeof(port), "%d", process_list[i]);
+					snprintf(mesg+total_port_len, strlen(port)+1, "%s", port);
+					total_port_len += strlen(port)+1;
 				}
 
 				pcp_write(frontend, "n", 1);
@@ -753,12 +751,91 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 				break;
 			}
 
+			case 'W':			/* watchdog info */
+			{
+				int wd_index;
+				int wsize;
+
+				WdInfo *wi = NULL;
+
+				if (!pool_config->use_watchdog)
+				{
+					char code[] = "watchdog not enabled";
+
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(code) + sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+
+					pool_debug("pcp_child: watcdhog not enabled");
+					break;
+				}
+
+				wd_index = atoi(buf);
+				wi = wd_get_watchdog_info(wd_index);
+
+				if (wi == NULL) {
+					char code[] = "Invalid watchdog index";
+
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(code) + sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+
+					pool_debug("pcp_child: invalid watchdog index");
+				}
+				else
+				{
+					char code[] = "CommandComplete";
+					char pgpool_port_str[6];
+					char wd_port_str[6];
+					char status[2];
+
+					snprintf(pgpool_port_str, sizeof(pgpool_port_str), "%d", wi->pgpool_port);
+					snprintf(wd_port_str, sizeof(wd_port_str), "%d", wi->wd_port);
+					snprintf(status, sizeof(status), "%d", wi->status);
+
+					pcp_write(frontend, "w", 1);
+					wsize = htonl(sizeof(code) +
+								  strlen(wi->hostname)+1 +
+								  strlen(pgpool_port_str)+1 +
+								  strlen(wd_port_str)+1 +
+								  strlen(status)+1 +
+								  sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+
+					pcp_write(frontend, wi->hostname, strlen(wi->hostname)+1);
+					pcp_write(frontend, pgpool_port_str, strlen(pgpool_port_str)+1);
+					pcp_write(frontend, wd_port_str, strlen(wd_port_str)+1);
+					pcp_write(frontend, status, strlen(status)+1);
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+
+					pool_debug("pcp_child: retrieved node information from shared memory");
+				}
+				break;
+			}
+
 			case 'D':			/* detach node */
 			case 'd':			/* detach node gracefully */
 			{
 				int node_id;
 				int wsize;
-				char *code;
+				char code[] = "CommandComplete";
 				bool gracefully;
 
 				if (tos == 'D')
@@ -768,15 +845,12 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 				node_id = atoi(buf);
 				pool_debug("pcp_child: detaching Node ID %d", node_id);
-				if (pool_detach_node(node_id, gracefully) == 0)
-					code = "CommandComplete";
-				else
-					code = "CommandFailed";
+				pool_detach_node(node_id, gracefully);
 
 				pcp_write(frontend, "d", 1);
-				wsize = htonl(strlen(code) + 1 + sizeof(int));
+				wsize = htonl(sizeof(code) + sizeof(int));
 				pcp_write(frontend, &wsize, sizeof(int));
-				pcp_write(frontend, code, strlen(code) + 1);
+				pcp_write(frontend, code, sizeof(code));
 				if (pcp_flush(frontend) < 0)
 				{
 					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
@@ -873,7 +947,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP))
 						msg = "primary server cannot be recovered by online recovery.";
 					else
-						msg = "recovery request is only allowed in replication and streaming replication modes.";
+						msg = "recovery request is accepted only in replication mode or stereaming replication mode. ";
 
 					len = strlen(msg)+1;
 					pcp_write(frontend, "e", 1);
@@ -1015,7 +1089,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					exit(1);
 				}
 
-				if (node_id == REAL_PRIMARY_NODE_ID)
+				if (node_id == PRIMARY_NODE_ID)
 				{
 					char code[] = "NodeIdAlreadyPrimary";
 					pool_error("pcp_child: specified node is already primary node, can't promote node id %d", node_id);
@@ -1328,26 +1402,12 @@ static RETSIGTYPE reload_config_handler(int sig)
 /* Dedatch a node */
 static int pool_detach_node(int node_id, bool gracefully)
 {
-	int nRet = 0;
 	if (!gracefully)
 	{
-		if (degenerate_backend_set_ex(&node_id, 1, false) == false)
-		{
-			pool_error("pcp_child: processing detach node failed");
-			return -1;
-		}
+		notice_backend_error(node_id);	/* send failover request */
 		return 0;
 	}
-
-	/* Check if the NODE DOWN can be executed on
-	 * the given node id.
-	 */
-	if (degenerate_backend_set_ex(&node_id, 1, true) == false)
-	{
-		pool_error("pcp_child: processing graceful detach node failed");
-		return -1;
-	}
-
+		
 	/*
 	 * Wait until all frontends exit
 	 */
@@ -1364,19 +1424,12 @@ static int pool_detach_node(int node_id, bool gracefully)
 	/*
 	 * Now all frontends have gone. Let's do failover.
 	 */
-	if (degenerate_backend_set_ex(&node_id, 1, false) == false)
-	{
-		nRet = -1;
-		pcp_wakeup_request = 1;
-		pool_error("pcp_child: processing graceful detach node failed");
-	}
-	else
-	{
-		/*
-		 * Wait for failover completed.
-		 */
-		pcp_wakeup_request = 0;
-	}
+	notice_backend_error(node_id);		/* send failover request */
+
+	/*
+	 * Wait for failover completed.
+	 */
+	pcp_wakeup_request = 0;
 
 	while (!pcp_wakeup_request)
 	{
@@ -1391,7 +1444,7 @@ static int pool_detach_node(int node_id, bool gracefully)
 	 */
 	finish_recovery();
 
-	return nRet;
+	return 0;
 }
 
 /* Promote a node */
@@ -1402,7 +1455,7 @@ static int pool_promote_node(int node_id, bool gracefully)
 		promote_backend(node_id);	/* send promote request */
 		return 0;
 	}
-
+		
 	/*
 	 * Wait until all frontends exit
 	 */
