@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2012	PgPool Global Development Group
+ * Copyright (c) 2003-2014	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -430,9 +430,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			int stime = 5;	/* XXX give arbitrary time to allow closing idle connections */
 
 			pool_debug("Query: sending SIGUSR1 signal to parent");
-
-			Req_info->kind = CLOSE_IDLE_REQUEST;
-			kill(getppid(), SIGUSR1);		/* send USR1 signal to parent */
+			register_node_operation_request(CLOSE_IDLE_REQUEST, NULL, 0);
 
 			/* we need to loop over here since we will get USR1 signal while sleeping */
 			while (stime > 0)
@@ -701,7 +699,9 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/*
 	 * Fetch memory cache if possible
 	 */
-	if (pool_config->memory_cache_enabled && pool_is_likely_select(query))
+	if (pool_config->memory_cache_enabled && pool_is_likely_select(query) &&
+		!pool_is_writing_transaction() &&
+		(TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != 'E'))
 	{
 		bool foundp;
 		POOL_STATUS status;
@@ -722,7 +722,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		/*
 		 * Add bind message's info to query to search.
 		 */
-		if (bind_msg->param_offset && bind_msg->contents)
+		if (query_context->is_cache_safe && bind_msg->param_offset && bind_msg->contents)
 		{
 			/* Extract binary contents from bind message */
 			char *query_in_bind_msg = bind_msg->contents + bind_msg->param_offset;
@@ -745,7 +745,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				snprintf(hex_str, sizeof(hex_str), (i == 0) ? " %02X" : "%02X", 0xff & query_in_bind_msg[i]);
 				hexlen = strlen(hex_str);
 
-				if ((len+hexlen) > alloc_len)
+				if ((len+hexlen) >= alloc_len)
 				{
 					alloc_len += STR_ALLOC_SIZE;
 					search_query = realloc(search_query, alloc_len);
@@ -1242,11 +1242,12 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 {
 	char *pstmt_name;
 	char *portal_name;
-	char *rewrite_msg;
+	char *rewrite_msg = NULL;
 	POOL_SENT_MESSAGE *parse_msg;
 	POOL_SENT_MESSAGE *bind_msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
+	int insert_stmt_with_lock = 0;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -1292,7 +1293,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	 */
 	if (query_context->is_cache_safe)
 	{
-		bind_msg->param_offset = sizeof(int) + sizeof(char) * (strlen(portal_name) + strlen(pstmt_name));
+		bind_msg->param_offset = sizeof(char) * (strlen(portal_name) + strlen(pstmt_name) + 2);
 	}
 
 	session_context->uncompleted_message = bind_msg;
@@ -1316,25 +1317,47 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			return POOL_END;
 	}
 
+	/*
+	 * Start a transaction if necessary in replication mode
+	 */
+	if (REPLICATION)
+	{
+		pool_debug("Bind: checking strict query");
+		if (is_strict_query(query_context->parse_tree))
+		{
+			pool_debug("Bind: strict query");
+			start_internal_transaction(frontend, backend, query_context->parse_tree);
+			allow_close_transaction = 1;
+		}
+
+		pool_debug("Bind: checking insert lock");
+		insert_stmt_with_lock = need_insert_lock(backend, query_context->original_query, query_context->parse_tree);
+		if (insert_stmt_with_lock)
+		{
+			pool_debug("Bind: issuing insert lock");
+			/* issue a LOCK command to keep consistency */
+			if (insert_lock(frontend, backend, query_context->original_query, (InsertStmt *)query_context->parse_tree, insert_stmt_with_lock) != POOL_CONTINUE)
+			{
+				pool_query_context_destroy(query_context);
+				return POOL_END;
+			}
+		}
+	}
+
 	pool_debug("Bind: waiting for master completing the query");
 	if (pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
-		if (rewrite_msg != NULL)
-			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
 
 	if (pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
-		if (rewrite_msg != NULL)
-			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
-
-	if (rewrite_msg != NULL)
-		/* free(rewrite_msg) */;
+    if(rewrite_msg)
+        free(rewrite_msg);
 
 	return POOL_CONTINUE;
 }
@@ -1510,7 +1533,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	int i;
 	int len;
 	signed char kind;
-	signed char state;
+	signed char state = 0;
 	POOL_SESSION_CONTEXT *session_context;
 	Node *node = NULL;
 	char *query = NULL;
@@ -1533,7 +1556,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	if (session_context->mismatch_ntuples && MAJOR(backend) == PROTO_MAJOR_V3)
 	{
 		int i;
-		signed char state;
 		char kind;
 
 		/*
@@ -1799,7 +1821,8 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				 */
 				if (pool_is_doing_extended_query_message())
 				{
-					if (session_context->query_context->query_state[MASTER_NODE_ID] == POOL_EXECUTE_COMPLETE)
+					if (session_context->query_context &&
+						session_context->query_context->query_state[MASTER_NODE_ID] == POOL_EXECUTE_COMPLETE)
 					{
 						pool_handle_query_cache(backend, session_context->query_context->query_w_hex, node, state);
 						free(session_context->query_context->query_w_hex);
@@ -1808,6 +1831,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				}
 				else
 				{
+					if (MAJOR(backend) != PROTO_MAJOR_V3)
+					{
+						state = 'I';	/* XXX I don't think query cache works with PROTO2 protocol */
+					}
 					pool_handle_query_cache(backend, query, node, state);
 				}
 			}
@@ -2231,6 +2258,111 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 	return POOL_CONTINUE;
 }
 
+POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
+								 POOL_CONNECTION_POOL *backend)
+{
+	int len, len1 = 0;
+	char *p = NULL;
+	char *p1 = NULL;
+	int status;
+	int sendlen;
+	int i;
+
+	POOL_SESSION_CONTEXT *session_context;
+	int num_params, send_num_params, num_dmy;
+	char kind = 't';
+
+	session_context = pool_get_session_context();
+	if (!session_context)
+	{
+		pool_error("ParameterDescription: cannot get session context");
+		return POOL_END;
+	}
+
+	/* only in replication mode and rewritten query */
+	if (!REPLICATION || !session_context->query_context->rewritten_query)
+		return SimpleForwardToFrontend('t', frontend, backend);
+
+	/* get number of parameters in original query */
+	num_params = session_context->query_context->num_original_params;
+
+	status = pool_read(MASTER(backend), &len, sizeof(len));
+	if (status < 0)
+	{
+		pool_error("ParameterDescription: error while reading message length");
+		return POOL_END;
+	}
+
+	len = ntohl(len);
+	len -= sizeof(int32);
+	len1 = len;
+
+	/* number of parameters in rewritten query is just discarded */
+	status = pool_read(MASTER(backend), &num_dmy, sizeof(int16));
+	len -= sizeof(int16);
+
+	p = pool_read2(MASTER(backend), len);
+	if (p == NULL)
+		return POOL_END;
+
+	p1 = malloc(len);
+	if (p1 == NULL)
+	{
+		pool_error("ParameterDescription: malloc failed");
+		return POOL_ERROR;
+	}
+	memcpy(p1, p, len);
+
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (VALID_BACKEND(i) && !IS_MASTER_NODE_ID(i))
+		{
+			status = pool_read(CONNECTION(backend, i), &len, sizeof(len));
+			if (status < 0)
+			{
+				pool_error("ParameterDescription: error while reading message length");
+                free(p1);
+				return POOL_END;
+			}
+
+			len = ntohl(len);
+			len -= sizeof(int32);
+
+			p = pool_read2(CONNECTION(backend, i), len);
+			if (p == NULL)
+            {
+                free(p1);
+				return POOL_END;
+            }
+			if (len != len1)
+			{
+				pool_debug("ParameterDescription: length does not match between backends master(%d) %d th backend(%d) kind:(%c)",
+						   len, i, len1, kind);
+			}
+		}
+	}
+
+	pool_write(frontend, &kind, 1);
+
+	/* send back OIDs of parameters in original query and left are discarded */
+	len = sizeof(int16) + num_params * sizeof(int32);
+	sendlen = htonl(len + sizeof(int32));
+	pool_write(frontend, &sendlen, sizeof(int32));
+
+	send_num_params = htons(num_params);
+	pool_write(frontend, &send_num_params, sizeof(int16));
+
+	if (pool_write_and_flush(frontend, p1, num_params * sizeof(int32)) < 0)
+	{
+		pool_error("ParameterDescription: pool_write_and_flush failed");
+		free(p1);
+		return POOL_END;
+	}
+
+	free(p1);
+	return POOL_CONTINUE;
+}
+
 POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
@@ -2413,7 +2545,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	session_context = pool_get_session_context();
 	if (!session_context)
 	{
-		pool_error("Parse: cannot get session context");
+		pool_error("ProcessFrontendResponse: cannot get session context");
 		return POOL_END;
 	}
 
@@ -2423,7 +2555,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
 		pool_log("ProcessFrontendResponse: failed to read kind from frontend. frontend abnormally exited");
-		return POOL_END;
+		return POOL_END_WITH_FRONTEND_ERROR;
 	}
 
 	pool_debug("ProcessFrontendResponse: kind from frontend %c(%02x)", fkind, fkind);
@@ -2680,6 +2812,10 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 					pool_unset_query_in_progress();
 				break;
 
+			case 't':	/* ParameterDescription */
+				status = ParameterDescription(frontend, backend);
+				break;
+
 			case 'I':	/* EmptyQueryResponse */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
 				if (pool_is_doing_extended_query_message())
@@ -2916,8 +3052,9 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 						id = pool_get_id(info, p1);
 						pool_debug("CopyDataRow: copying id: %d", id);
 						free(p1);
-						if (!VALID_BACKEND(id))
+						if (id < 0 || !VALID_BACKEND(id))
 						{
+							pool_error("CopyDataRow: pool_get_id returns invalid id: %d", id);
 							exit(1);
 						}
 						if (pool_write(CONNECTION(backend, id), &kind, 1))
